#include "Vector.hpp"
#include "Voronoi.hpp"
#include "Polygon.hpp"

#include <stdio.h>
#include <dos.h>
#include <conio.h>
#include <assert.h>
#include <values.h>
#include <graphics.h>
#include <math.h>
#include <stdlib.h>
#include "mouse.hpp"

#define DEBUG_DRAW


//#define ABS(X) ((X)<0?(-(X)):(X))
#define BEEP(TONE,DELAY)   { sound(TONE); delay(DELAY); nosound(); }

#define MAX(A,B) ((A)>(B)?(A):(B))



//#define img_dx  (368/2*8)
//#define img_dy  (280/2*8)

Voronoi::Voronoi(void) {
}


Voronoi::~Voronoi() {
}



// This function calls TryCalcAllPol().
// If TryCalcAllPol() fails the point for which it failed will be
// slightly changed and try again.
int Voronoi::CalcAllPol(ArrayP2d& a_point, Polygon* apol) {
    int wrong_p;
	int result;

	int count = 20;
	Point2d org_point;

	result = TryCalcAllPol(a_point, apol, &wrong_p);
	if(result==NULL) org_point = a_point[wrong_p];
	while (result == NULL) {

		a_point[wrong_p] = org_point + Point2d(random(3)-1, random(3)-1);
		result = TryCalcAllPol(a_point, apol, &wrong_p, wrong_p);
		if(count-- < 0) return 0;  // ERROR

	}
	return 1; // OK

}



int Voronoi::GetNeighboursList(
     int nCentralPoint,
     int nNumberOfPoints,
     Polygon *pPol,
     int nMaxNeighbour,
     int *pnNeighbourList,       // filled by function
     int *pnNeighbourCount)     // filled by function
{

	*pnNeighbourCount = 0;

    for(int i=0; i<nNumberOfPoints; i++) {
		if(i!=nCentralPoint) {

			if(AreNeighbours(pPol[i], pPol[nCentralPoint])) {
                if(*pnNeighbourCount>=nMaxNeighbour) return 0; // list size not enough
                pnNeighbourList[*pnNeighbourCount] = i;
            	(*pnNeighbourCount)++;
            }
        }
    }
	return 1;
}




int Voronoi::TryCalcAllPol(ArrayP2d& a_point, Polygon* apol, int* wrong_p, int start_with/* =0 */) {
	int i;

	for(i=start_with; i<a_point.Count(); i++) {
		if(CalcPol(a_point, apol[i], i) == NULL) {
			*wrong_p = i;
			return NULL;
		}
	}

	for(i=0; i<start_with; i++) {
		if(CalcPol(a_point, apol[i], i) == NULL) {
			*wrong_p = i;
			return NULL;
		}
	}

	return 1;   // OK

}



int Voronoi::CalcPol(ArrayP2d& a_point, Polygon& pol, int i_p) {
//	int lim_x;
//	int lim_y;

	int i;

	// repeated points are not allowed
	for(i=0; i<a_point.Count(); i++) {
		if(i!= i_p) if(a_point[i] == a_point[i_p]) {
//			BEEP(4000,50);
			return NULL;
		}

	}

//	lim_x = lim_y = 2*MAX(img_dx, img_dy);

	pol.Clear();
/*	pol.AddPoint((a_point[i_p] + Point2d(-lim_x/2,+lim_y/2)));
	pol.AddPoint(a_point[i_p] + Point2d(+lim_x/2,+lim_y/2));
	pol.AddPoint(a_point[i_p] + Point2d(+lim_x/2,-lim_y/2));
	pol.AddPoint(a_point[i_p] + Point2d(-lim_x/2,-lim_y/2));
*/
	pol.AddPoint((a_point[i_p] + Point2d(-img_dx/2,+img_dy/2)));
	pol.AddPoint(a_point[i_p] + Point2d(+img_dx/2,+img_dy/2));
	pol.AddPoint(a_point[i_p] + Point2d(+img_dx/2,-img_dy/2));
	pol.AddPoint(a_point[i_p] + Point2d(-img_dx/2,-img_dy/2));


	Segment seg_med;
	for(i=0; i<a_point.Count(); i++) {
		if(i!=i_p) {
			seg_med = GetMidSegment(a_point[i], a_point[i_p]);
			//setcolor(BLUE);
			//seg_med.Draw();

			// checks the point is close enough to possibly clip the polygon
			if(ABS(a_point[i].x-a_point[i_p].x) <= img_dx &&
			   ABS(a_point[i].y-a_point[i_p].y) <= img_dy) {

				if(pol.Clip(seg_med, a_point[i_p]) == NULL) {
					BEEP(2000,20);
					return NULL;
				}

			}

		}
	}


	return 1; // OK
}







void Voronoi::Render(ArrayP2d& a_point, int resolution) {

	int x,y, i,j;
	long l_dist2;
	long l_min_dist2 = MAXLONG;
	int  i_min_index=0;
	long dx,dy;

	int org_x=0;
	int org_i_min_index=-1;

	setcolor(i_min_index);

	for(y=0; y<480; y+=resolution) {
		for(x=0; x<640; x+=resolution) {
			l_min_dist2 = MAXLONG;
			i_min_index=-1;
			for(i=0; i<a_point.Count(); i++) {
				dx = ABS(a_point[i].x-x);
				dy = ABS(a_point[i].y-y);
				if(dx<img_dx/2 && dy<img_dy/2) {
					l_dist2 = dx*dx + dy*dy;
					//if(l_dist2<100*100) {
						if(l_dist2 < l_min_dist2) {
							l_min_dist2 = l_dist2;
							i_min_index = i;
						}
					//}
				}
			}


			if(resolution != 1) {
				setfillstyle(SOLID_FILL, i_min_index+1);
				bar(x,y,x+resolution,y+resolution);
			} else {
				//putpixel(x,y, i_min_index+1);
				if(x==639 || i_min_index!=org_i_min_index) {
					line(org_x, y, x,y);
					if(x==639) org_x=0;
					else org_x=x;
					org_i_min_index = i_min_index;
					setcolor(i_min_index+1);
				}
			}

		}
	}
}





Segment Voronoi::GetMidSegment(Point2d& p1, Point2d& p2) {

	Vector mid;
	Vector dir;
	Vector perp;
	Vector seg_from;
	Vector seg_to;

	Vector v1(p1.x, p1.y, 0);
	Vector v2(p2.x, p2.y, 0);


	mid = v1 + (v2 - v1)/2.0;
	dir = (v2-v1);
	perp.x = -dir.y;
	perp.y = dir.x;

	perp.normaliza();

	seg_from = mid + perp*550;
	seg_to = mid - perp*550;

	return Segment(seg_from, seg_to);
}



/*
// neighbours share at least one vertex of their polygons
int Voronoi::AreNeighbours(Polygon &Pol1, Polygon &Pol2) {
	Point2d p2dI1I2;
    #define THRESHOLD 6

	for(int i1=0; i1<Pol1.point.Count(); i1++) {
   		for(int i2=0; i2<Pol2.point.Count(); i2++) {

        	p2dI1I2 = Pol1.point[i1] - Pol2.point[i2];
	        if(p2dI1I2.mod()<2.0) return 1;
	        if(p2dI1I2.mod()<THRESHOLD) {

                int i1i = Pol1.IncIndex(i1);
	            int i1d = Pol1.DecIndex(i1);

	            int i2i = Pol2.IncIndex(i2);
	            int i2d = Pol2.DecIndex(i2);

	            float fDii = (Pol1.point[i1i] - Pol2.point[i2i]).mod();
	            float fDdd = (Pol1.point[i1d] - Pol2.point[i2d]).mod();
	            float fDdi = (Pol1.point[i1d] - Pol2.point[i2i]).mod();
	            float fDid = (Pol1.point[i1i] - Pol2.point[i2d]).mod();

                int j1,j2;
				if(              fDii<=fDdd && fDii<=fDdi && fDii<=fDid) {j1=i1i; j2=i2i; }
				if(fDdd<=fDii               && fDdd<=fDdi && fDdd<=fDid) {j1=i1d; j2=i2d; }
				if(fDdi<=fDii && fDdi<=fDdd &&               fDdi<=fDid) {j1=i1d; j2=i2i; }
				if(fDid<=fDii && fDid<=fDdd && fDid<=fDdi              ) {j1=i1i; j2=i2d; }

				//Point2d p2dJ1J2 = Pol1.point[j1] - Pol2.point[j2];

				Point2d p2dI1J1 = Pol1.point[i1] - Pol1.point[j1];
				Point2d p2dI2J2 = Pol2.point[i2] - Pol2.point[j2];

				//Point2d p2dMidI1J1 = (Pol1.point[i1] - Pol1.point[j1])/2 + Pol1.point[j1];
				//Point2d p2dMidI2J2 = (Pol2.point[i2] - Pol2.point[j2])/2 + Pol2.point[j2];
                //if((p2dMidI1J1 - p2dMidI2J2).mod()<THRESHOLD) return 1;

				float fAtanI = atan2((float)(p2dI1J1.y), (float)(p2dI1J1.x));
  				float fAtanJ = atan2((float)(p2dI2J2.y), (float)(p2dI2J2.x));

                #define VALID_ANGLE_RANGE .1
                #define PI 3.14159
                if(ABS(fAtanI-fAtanJ) < 0+VALID_ANGLE_RANGE) return 1;
                if(ABS(fAtanI-fAtanJ) > 2*PI-VALID_ANGLE_RANGE) return 1;

            }


    	}
    }
    return 0;
}
*/


// 2nd Version: neighbours share one side of their polygons
int Voronoi::AreNeighbours(Polygon &Pol1, Polygon &Pol2) {
	Point2d p2dI1I2;
    #define THRESHOLD 6
    Segment s1,s2;
    Vector vMs1, vMs2;
    char buffer[300];


    setcolor(GREEN);

	for(int i1=0; i1<Pol1.point.Count(); i1++) {
   		for(int i2=0; i2<Pol2.point.Count(); i2++) {

        	s1.p1.x = Pol1.point[i1].x;
        	s1.p1.y = Pol1.point[i1].y;
        	s1.p2.x = Pol1.point[Pol1.IncIndex(i1)].x;
        	s1.p2.y = Pol1.point[Pol1.IncIndex(i1)].y;

        	s2.p1.x = Pol2.point[i2].x;
        	s2.p1.y = Pol2.point[i2].y;
        	s2.p2.x = Pol2.point[Pol2.IncIndex(i2)].x;
        	s2.p2.y = Pol2.point[Pol2.IncIndex(i2)].y;

            vMs1 = (s1.p1 + s1.p2) / 2.0f;
            vMs2 = (s2.p1 + s2.p2) / 2.0f;

            //sprintf(buffer, "vMs1(%f,%f,%f)", vMs1.x, vMs1.y, vMs1.z);
            //outtextxy(0, m_ny+=16, buffer);
            //sprintf(buffer, "vMs2(%f,%f,%f)", vMs2.x, vMs2.y, vMs2.z);
            //outtextxy(0, m_ny+=16, buffer);

			float fAtan1 = atan2((float)(s1.p2.y - s1.p1.y), (float)(s1.p2.x - s1.p1.x));
			float fAtan2 = atan2((float)(s2.p2.y - s2.p1.y), (float)(s2.p2.x - s2.p1.x));

            //setcolor(GREEN);


            if( s1.IsInSegment(s1.GetNearestPoint(s2.p1)) ||
	            s1.IsInSegment(s1.GetNearestPoint(s2.p2)) ||
	            s2.IsInSegment(s2.GetNearestPoint(s1.p1)) ||
	            s2.IsInSegment(s2.GetNearestPoint(s1.p2))) {

	            if(s1.Distance2(vMs2) < THRESHOLD || s2.Distance2(vMs1) < THRESHOLD ) {
		            #define VALID_ANGLE_RANGE .1
		            #define PI 3.14159
		            if(ABS(fAtan1-fAtan2) < VALID_ANGLE_RANGE) {/*setcolor(LIGHTGREEN); */return 1; }
		            if(ABS(ABS(fAtan1-fAtan2)-PI) < VALID_ANGLE_RANGE) {/*setcolor(LIGHTRED); */return 1; }
	            }
			}

//            sprintf(buffer, "s1-vMs2=%3.3f  s2-vMs1=%3.3f Atan1=%4.4f Atan2=%4.4f",
//				s1.Distance2(vMs2), s2.Distance2(vMs1), fAtan1, fAtan2);
//            outtextxy(0, m_ny+=16, buffer);

    	}
    }
    return 0;
}

